#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from functools import WRAPPER_ASSIGNMENTS, wraps


def ignore_unhashable(func):
    uncached = func.__wrapped__
    attributes = WRAPPER_ASSIGNMENTS + ("cache_info", "cache_clear")
    wraps(func, assigned=attributes)

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except TypeError as error:
            if "unhashable type" in str(error):
                return uncached(*args, **kwargs)
            raise

    wrapper.__uncached__ = uncached
    return wrapper



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Dict, Optional

from ethtx.models.base_model import BaseModel


class TransformationSemantics(BaseModel):
    transformed_name: Optional[str]
    transformed_type: Optional[str]
    transformation: str = ""


class ParameterSemantics(BaseModel):
    parameter_name: str
    parameter_type: str
    components: list = []
    indexed: bool = False
    dynamic: bool = False


class EventSemantics(BaseModel):
    signature: str
    anonymous: bool
    name: str
    parameters: List[ParameterSemantics]


class FunctionSemantics(BaseModel):
    signature: str
    name: str
    inputs: List[ParameterSemantics]
    outputs: List[ParameterSemantics] = []


class SignatureArg(BaseModel):
    name: str
    type: str


class Signature(BaseModel):
    signature_hash: str
    name: str
    args: List[SignatureArg]
    count: int = 1
    tuple: bool = False
    guessed: bool = False


class ERC20Semantics(BaseModel):
    name: str
    symbol: str
    decimals: int


class ContractSemantics(BaseModel):
    code_hash: str
    name: str
    events: Dict[str, EventSemantics] = {}
    functions: Dict[str, FunctionSemantics] = {}
    transformations: Dict[str, Dict[str, TransformationSemantics]] = {}


class AddressSemantics(BaseModel):
    chain_id: str
    address: str
    name: str
    is_contract: bool
    contract: ContractSemantics
    standard: Optional[str]
    erc20: Optional[ERC20Semantics]

    class Config:
        allow_mutation = True



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from ethtx.models.semantics_model import FunctionSemantics, ParameterSemantics

precompiles = {
    1: FunctionSemantics(
        signature="",
        name="ecrecover",
        inputs=[
            ParameterSemantics(parameter_name="hash", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="v", parameter_type="bytes8"),
            ParameterSemantics(parameter_name="r", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="s", parameter_type="bytes32"),
        ],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="address")],
    ),
    2: FunctionSemantics(
        signature="",
        name="sha256",
        inputs=[ParameterSemantics(parameter_name="data", parameter_type="raw")],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="bytes32")],
    ),
    3: FunctionSemantics(
        signature="",
        name="ripemd160",
        inputs=[ParameterSemantics(parameter_name="data", parameter_type="raw")],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="bytes32")],
    ),
    4: FunctionSemantics(
        signature="",
        name="datacopy",
        inputs=[ParameterSemantics(parameter_name="data", parameter_type="raw")],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="raw")],
    ),
    5: FunctionSemantics(
        signature="",
        name="bigModExp",
        inputs=[
            ParameterSemantics(parameter_name="base", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="exp", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="mod", parameter_type="bytes32"),
        ],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="bytes32")],
    ),
    6: FunctionSemantics(
        signature="",
        name="bn256Add",
        inputs=[
            ParameterSemantics(parameter_name="ax", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="ay", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="bx", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="by", parameter_type="bytes32"),
        ],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="bytes32[2]")],
    ),
    7: FunctionSemantics(
        signature="",
        name="bn256ScalarMul",
        inputs=[
            ParameterSemantics(parameter_name="x", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="y", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="scalar", parameter_type="bytes32"),
        ],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="bytes32[2]")],
    ),
    8: FunctionSemantics(
        signature="",
        name="bn256Pairing",
        inputs=[ParameterSemantics(parameter_name="input", parameter_type="raw")],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="bytes32")],
    ),
}



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from abc import ABC, abstractmethod
from typing import Any, TypeVar, Type

from ens import ENS
from web3 import Web3
from web3.middleware import geth_poa_middleware

log = logging.getLogger(__name__)

T = TypeVar("T")


class ENSProviderBase(ABC):
    @abstractmethod
    def name(self, provider: Type[T], address: Any):
        ...

    @abstractmethod
    def address(self, provider: Type[T], name: Any):
        ...


class Web3ENSProvider(ENSProviderBase):
    ns: ENS

    def name(self, provider: Web3, address: str) -> str:
        ns = self._set_provider(provider)
        check_sum_address = Web3.toChecksumAddress(address)
        name = ns.name(address=check_sum_address)

        if name:
            log.info("ENS resolved an address: %s to name: %s", address, name)

        return name if name else address

    def address(self, provider: Web3, name: str) -> str:
        ns = self._set_provider(provider)
        address = ns.address(name=name)

        if address:
            log.info("ENS resolved name: %s to address: %s", name, address)

        return address if address else name

    @staticmethod
    def _set_provider(provider: Web3) -> ENS:
        ns = ENS.fromWeb3(provider)

        # ENS.fromWeb3 not copying middleware #1657
        ns.web3.middleware_onion.inject(geth_poa_middleware, layer=0)

        return ns


ENSProvider = Web3ENSProvider()



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Dict, List

from .connection_base import NodeConnection


class NodeConnectionPool:
    def __init__(self, nodes: Dict[str, dict]):
        self._connections: List[NodeConnection] = []

        self._set_connections(nodes)

    def __len__(self) -> int:
        return len(self._connections)

    @property
    def connections(self) -> List[NodeConnection]:
        return self._connections

    def add_connection(self, connection: NodeConnection) -> None:
        if not isinstance(connection, NodeConnection):
            raise ValueError("Value is not instance of NodeBase")

        self._connections.append(connection)

    def get_connection(self, chain: str) -> List[NodeConnection]:
        return [
            connection for connection in self._connections if connection.chain == chain
        ]

    def _set_connections(self, nodes) -> None:
        for chain, node_params in nodes.items():
            nodes: List[str] = list(node_params.values())[0].split(",")
            poa: bool = list(node_params.values())[1]
            for url in nodes:
                self.add_connection(
                    NodeConnection(chain=chain, url=url.strip(), poa=poa)
                )



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from functools import lru_cache
from typing import Optional, List, Dict, Tuple

from ethtx.decoders.decoders.semantics import decode_events_and_functions
from ethtx.models.semantics_model import (
    AddressSemantics,
    ContractSemantics,
    ParameterSemantics,
    ERC20Semantics,
    TransformationSemantics,
    FunctionSemantics,
    EventSemantics,
    Signature,
    SignatureArg,
)
from ethtx.providers import EtherscanProvider, Web3Provider, ENSProvider
from ethtx.providers.semantic_providers.database import ISemanticsDatabase
from ethtx.semantics.protocols_router import amend_contract_semantics
from ethtx.semantics.solidity.precompiles import precompiles
from ethtx.semantics.standards.erc20 import ERC20_FUNCTIONS, ERC20_EVENTS
from ethtx.semantics.standards.erc721 import ERC721_FUNCTIONS, ERC721_EVENTS


class SemanticsRepository:
    def __init__(
        self,
        database_connection: ISemanticsDatabase,
        etherscan_provider: EtherscanProvider,
        web3provider: Web3Provider,
        ens_provider: ENSProvider,
    ):
        self.database = database_connection
        self.etherscan = etherscan_provider
        self._web3provider = web3provider
        self._ens_provider = ens_provider

        self._records: Optional[List] = None

    def record(self) -> None:
        """Records is an array used to hold semantics used in tx decing process.
        This recording is used just for logging"""
        self._records = []

    def end_record(self) -> List:
        tmp_records = self._records
        self._records = None
        return tmp_records

    def _read_stored_semantics(
        self, address: str, chain_id: str
    ) -> Optional[AddressSemantics]:
        def decode_parameter(_parameter):
            components_semantics = []
            if "components" in _parameter:
                for component in _parameter["components"]:
                    components_semantics.append(decode_parameter(component))

            decoded_parameter = ParameterSemantics(
                parameter_name=_parameter["parameter_name"],
                parameter_type=_parameter["parameter_type"],
                components=components_semantics,
                indexed=_parameter["indexed"],
                dynamic=_parameter["dynamic"],
            )

            return decoded_parameter

        if not address:
            return None

        ZERO_HASH = "0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470"

        raw_address_semantics = self.database.get_address_semantics(chain_id, address)

        if raw_address_semantics:

            if raw_address_semantics.get("erc20"):
                erc20_semantics = ERC20Semantics(
                    name=raw_address_semantics["erc20"]["name"],
                    symbol=raw_address_semantics["erc20"]["symbol"],
                    decimals=raw_address_semantics["erc20"]["decimals"],
                )
            else:
                erc20_semantics = None

            if raw_address_semantics["contract"] == ZERO_HASH:
                contract_semantics = ContractSemantics(
                    code_hash=raw_address_semantics["contract"], name="EOA"
                )

            else:

                raw_contract_semantics = self.database.get_contract_semantics(
                    raw_address_semantics["contract"]
                )
                events = {}

                for signature, event in raw_contract_semantics["events"].items():

                    parameters_semantics = []
                    for parameter in event["parameters"]:
                        parameters_semantics.append(decode_parameter(parameter))

                    events[signature] = EventSemantics(
                        signature=signature,
                        anonymous=event["anonymous"],
                        name=event["name"],
                        parameters=parameters_semantics,
                    )

                functions = {}
                for signature, function in raw_contract_semantics["functions"].items():

                    inputs_semantics = []
                    for parameter in function["inputs"]:
                        inputs_semantics.append(decode_parameter(parameter))
                    outputs_semantics = []
                    for parameter in function["outputs"]:
                        outputs_semantics.append(decode_parameter(parameter))

                    functions[signature] = FunctionSemantics(
                        signature=signature,
                        name=function["name"],
                        inputs=inputs_semantics,
                        outputs=outputs_semantics,
                    )

                transformations = {}
                for signature, parameters_transformations in raw_contract_semantics[
                    "transformations"
                ].items():
                    transformations[signature] = {}
                    for parameter, transformation in parameters_transformations.items():
                        transformations[signature][parameter] = TransformationSemantics(
                            transformed_name=transformation["transformed_name"],
                            transformed_type=transformation["transformed_type"],
                            transformation=transformation["transformation"],
                        )

                contract_semantics = ContractSemantics(
                    code_hash=raw_contract_semantics["code_hash"],
                    name=raw_contract_semantics["name"],
                    events=events,
                    functions=functions,
                    transformations=transformations,
                )

            name = raw_address_semantics.get("name", address)
            if name == address and not raw_address_semantics["is_contract"]:
                name = self._ens_provider.name(
                    provider=self._web3provider._get_node_connection(chain_id),
                    address=address,
                )

            address_semantics = AddressSemantics(
                chain_id=chain_id,
                address=address,
                name=name,
                is_contract=raw_address_semantics["is_contract"],
                contract=contract_semantics,
                standard=raw_address_semantics["standard"],
                erc20=erc20_semantics,
            )

            return address_semantics

        return None

    @lru_cache(maxsize=128)
    def get_semantics(self, chain_id: str, address: str) -> Optional[AddressSemantics]:

        if not address:
            return None

        ZERO_HASH = "0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470"

        address_semantics = self._read_stored_semantics(address, chain_id)
        if not address_semantics:

            # try to read the semantics form the Etherscan provider
            provider = self._web3provider
            code_hash = provider.get_code_hash(address, chain_id)

            if code_hash != ZERO_HASH:
                # smart contract
                raw_semantics, decoded = self.etherscan.contract.get_contract_abi(
                    chain_id, address
                )
                if decoded and raw_semantics:
                    # raw semantics received from Etherscan
                    events, functions = decode_events_and_functions(
                        raw_semantics["abi"]
                    )
                    standard, standard_semantics = self._decode_standard_semantics(
                        address, raw_semantics["name"], events, functions
                    )
                    if standard == "ERC20":
                        erc20_semantics = standard_semantics
                    else:
                        proxy_erc20 = provider.guess_erc20_proxy(address, chain_id)
                        if proxy_erc20:
                            erc20_semantics = ERC20Semantics(**proxy_erc20)
                        else:
                            erc20_semantics = None
                    contract_semantics = ContractSemantics(
                        code_hash=code_hash,
                        name=raw_semantics["name"],
                        events=events,
                        functions=functions,
                        transformations={},
                    )
                    address_semantics = AddressSemantics(
                        chain_id=chain_id,
                        address=address,
                        name=raw_semantics["name"],
                        is_contract=True,
                        contract=contract_semantics,
                        standard=standard,
                        erc20=erc20_semantics,
                    )

                else:
                    # try to guess if the address is a toke
                    potential_erc20_semantics = provider.guess_erc20_token(
                        address, chain_id
                    )
                    if potential_erc20_semantics:
                        standard = "ERC20"
                        erc20_semantics = ERC20Semantics(
                            name=potential_erc20_semantics["name"],
                            symbol=potential_erc20_semantics["symbol"],
                            decimals=potential_erc20_semantics["decimals"],
                        )
                    else:
                        standard = None
                        erc20_semantics = None

                    contract_semantics = ContractSemantics(
                        code_hash=code_hash, name=address
                    )
                    address_semantics = AddressSemantics(
                        chain_id=chain_id,
                        address=address,
                        name=address,
                        is_contract=True,
                        contract=contract_semantics,
                        standard=standard,
                        erc20=erc20_semantics,
                    )

            else:
                # externally owned address
                contract_semantics = ContractSemantics(code_hash=ZERO_HASH, name="EOA")
                name = self._ens_provider.name(
                    provider=self._web3provider._get_node_connection(chain_id),
                    address=address,
                )
                address_semantics = AddressSemantics(
                    chain_id=chain_id,
                    address=address,
                    name=name,
                    is_contract=False,
                    contract=contract_semantics,
                )

            self.update_semantics(address_semantics)

        # amend semantics with locally stored updates
        amend_contract_semantics(address_semantics.contract)

        if self._records is not None:
            self._records.append(address)

        return address_semantics

    def _decode_standard_semantics(
        self, address, name, events, functions
    ) -> Tuple[Optional[str], Optional[ERC20Semantics]]:
        standard = None
        standard_semantics = None

        if not address:
            return standard, standard_semantics

        if all(erc20_event in events for erc20_event in ERC20_EVENTS) and all(
            erc20_function in functions for erc20_function in ERC20_FUNCTIONS
        ):
            standard = "ERC20"
            try:
                provider = self._web3provider
                token_data = provider.get_erc20_token(address, name, functions)
                standard_semantics = ERC20Semantics(
                    name=token_data["name"],
                    symbol=token_data["symbol"],
                    decimals=token_data["decimals"],
                )
            except Exception:
                standard_semantics = ERC20Semantics(name=name, symbol=name, decimals=18)
        elif all(erc721_event in events for erc721_event in ERC721_EVENTS) and all(
            erc721_function in functions for erc721_function in ERC721_FUNCTIONS
        ):
            standard = "ERC721"
            standard_semantics = None

        return standard, standard_semantics

    @lru_cache(maxsize=128)
    def get_event_abi(self, chain_id, address, signature) -> Optional[EventSemantics]:

        if not address:
            return None

        semantics = self.get_semantics(chain_id, address)
        event_semantics = (
            semantics.contract.events.get(signature) if semantics else None
        )

        return event_semantics

    @lru_cache(maxsize=128)
    def get_transformations(
        self, chain_id, address, signature
    ) -> Optional[Dict[str, TransformationSemantics]]:

        if not address:
            return None

        semantics = self.get_semantics(chain_id, address)
        if semantics:
            transformations = semantics.contract.transformations.get(signature)
        else:
            transformations = None

        return transformations

    @lru_cache(maxsize=128)
    def get_anonymous_event_abi(self, chain_id, address) -> Optional[EventSemantics]:

        if not address:
            return None

        semantics = self.get_semantics(chain_id, address)
        event_semantics = None
        if semantics:
            anonymous_events = {
                signature
                for signature, event in semantics.contract.events.items()
                if event.anonymous
            }
            if len(anonymous_events) == 1:
                event_signature = anonymous_events.pop()
                event_semantics = semantics.contract.events[event_signature]

        return event_semantics

    @lru_cache(maxsize=128)
    def get_function_abi(
        self, chain_id, address, signature
    ) -> Optional[FunctionSemantics]:

        if not address:
            return None

        semantics = self.get_semantics(chain_id, address)
        function_semantics = (
            semantics.contract.functions.get(signature) if semantics else None
        )

        return function_semantics

    @lru_cache(maxsize=128)
    def get_constructor_abi(self, chain_id, address) -> Optional[FunctionSemantics]:

        if not address:
            return None

        semantics = self.get_semantics(chain_id, address)
        constructor_semantics = (
            semantics.contract.functions.get("constructor") if semantics else None
        )
        if constructor_semantics:
            constructor_semantics.outputs.append(
                ParameterSemantics(
                    parameter_name="__create_output__",
                    parameter_type="ignore",
                    indexed=False,
                    dynamic=True,
                )
            )

        return constructor_semantics

    def get_address_label(self, chain_id, address, proxies=None) -> str:

        if not address:
            return ""

        if int(address, 16) in precompiles:
            contract_label = "Precompiled"
        else:
            semantics = self.get_semantics(chain_id, address)
            if semantics.erc20:
                contract_label = semantics.erc20.symbol
            elif proxies and address in proxies:
                contract_label = proxies[address].name
            else:
                contract_label = (
                    semantics.name if semantics and semantics.name else address
                )

        return contract_label

    @lru_cache(maxsize=128)
    def check_is_contract(self, chain_id, address) -> bool:

        if not address:
            return False

        semantics = self.get_semantics(chain_id, address)
        is_contract = semantics is not None and semantics.is_contract

        return is_contract

    @lru_cache(maxsize=128)
    def get_standard(self, chain_id, address) -> Optional[str]:

        if not address:
            return None

        semantics = self.get_semantics(chain_id, address)
        standard = semantics.standard if semantics is not None else None

        return standard

    def get_token_data(
        self, chain_id, address, proxies=None
    ) -> Tuple[Optional[str], Optional[str], Optional[int], Optional[str]]:

        if not address:
            return None, None, None, None

        semantics = self.get_semantics(chain_id, address)
        if semantics and semantics.erc20:
            token_name = (
                semantics.erc20.name if semantics and semantics.erc20 else address
            )
            token_symbol = (
                semantics.erc20.symbol if semantics and semantics.erc20 else "Unknown"
            )
            token_decimals = (
                semantics.erc20.decimals if semantics and semantics.erc20 else 18
            )
        elif proxies and address in proxies and proxies[address].token:
            token_name = proxies[address].token.name
            token_symbol = proxies[address].token.symbol
            token_decimals = proxies[address].token.decimals
        else:
            token_name = address
            token_symbol = "Unknown"
            token_decimals = 18

        return token_name, token_symbol, token_decimals, "ERC20"

    def update_address(self, chain_id, address, contract) -> Dict:

        updated_address = {"network": chain_id, "address": address, **contract}
        self.database.insert_address(address=updated_address, update_if_exist=True)

        return updated_address

    def update_semantics(self, semantics) -> None:

        if not semantics:
            return

        contract_id = self.database.insert_contract(
            contract=semantics.contract.dict(), update_if_exist=True
        )

        updated_address_semantics = semantics.copy()
        updated_address_semantics.contract = (
            updated_address_semantics.contract.code_hash
        )
        self.database.insert_address(
            address=updated_address_semantics.dict(), update_if_exist=True
        )

        if contract_id:
            self.insert_contract_signatures(semantics.contract)

    def insert_contract_signatures(self, contract_semantics: ContractSemantics) -> None:
        for _, v in contract_semantics.functions.items():

            if not v.signature.startswith("0x"):
                continue

            if v.inputs and v.inputs[0].parameter_type == "tuple":
                args = [
                    SignatureArg(name=param.parameter_name, type=param.parameter_type)
                    for param in v.inputs[0].components
                ]
            else:
                args = (
                    [
                        SignatureArg(
                            name=param.parameter_name, type=param.parameter_type
                        )
                        for param in v.inputs
                    ]
                    if v.inputs
                    else []
                )

            new_signature = Signature(
                signature_hash=v.signature, name=v.name, args=args
            )

            self.update_or_insert_signature(new_signature)

    def get_most_used_signature(self, signature_hash: str) -> Optional[Signature]:
        signatures = list(
            self.database.get_signature_semantics(signature_hash=signature_hash)
        )

        if signatures:
            most_common_signature = max(signatures, key=lambda x: x["count"])
            signature = Signature(
                signature_hash=most_common_signature["signature_hash"],
                name=most_common_signature["name"],
                args=most_common_signature["args"],
                count=most_common_signature["count"],
                tuple=most_common_signature["tuple"],
                guessed=most_common_signature["guessed"],
            )

            return signature

        return None

    def update_or_insert_signature(self, signature: Signature) -> None:
        signatures = self.database.get_signature_semantics(
            signature_hash=signature.signature_hash
        )
        for sig in signatures:
            if (
                signature.name == sig["name"]
                and signature.signature_hash == sig["signature_hash"]
                and len(signature.args) == len(sig["args"])
            ):
                if signature.args and any(
                    arg for arg in list(sig["args"][0].values()) if "arg" in arg
                ):
                    for index, argument in enumerate(sig["args"]):
                        argument["name"] = signature.args[index].name
                        argument["type"] = signature.args[index].type

                sig["count"] += 1
                sig["guessed"] = False
                self.database.insert_signature(signature=sig, update_if_exist=True)
                break

        else:
            self.database.insert_signature(signature=signature.dict())



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ethtx.models.semantics_model import (
    EventSemantics,
    FunctionSemantics,
    ParameterSemantics,
    TransformationSemantics,
)

erc721_transfer_event = EventSemantics(
    signature="0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef",
    anonymous=False,
    name="Transfer",
    parameters=[
        ParameterSemantics(
            parameter_name="from", parameter_type="address", components=[], indexed=True
        ),
        ParameterSemantics(
            parameter_name="to", parameter_type="address", components=[], indexed=True
        ),
        ParameterSemantics(
            parameter_name="tokenId",
            parameter_type="uint256",
            components=[],
            indexed=True,
        ),
    ],
)

erc721_transfer_event_transformation = {
    "__input2__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input2__)"
    )
}

erc721_approval_event = EventSemantics(
    signature="0x8c5be1e5ebec7d5bd14f71427d1e84f3dd0314c0f7b2291e5b200ac8c7c3b925",
    anonymous=False,
    name="Approval",
    parameters=[
        ParameterSemantics(
            parameter_name="owner",
            parameter_type="address",
            components=[],
            indexed=True,
        ),
        ParameterSemantics(
            parameter_name="approved",
            parameter_type="address",
            components=[],
            indexed=True,
        ),
        ParameterSemantics(
            parameter_name="tokenId",
            parameter_type="uint256",
            components=[],
            indexed=True,
        ),
    ],
)

erc721_approval_event_transformation = {
    "__input2__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input2__)"
    )
}

erc721_approvalForAll_event = EventSemantics(
    signature="0x17307eab39ab6107e8899845ad3d59bd9653f200f220920489ca2b5937696c31",
    anonymous=False,
    name="ApprovalForAll",
    parameters=[
        ParameterSemantics(
            parameter_name="owner",
            parameter_type="address",
            components=[],
            indexed=True,
        ),
        ParameterSemantics(
            parameter_name="operator",
            parameter_type="address",
            components=[],
            indexed=True,
        ),
        ParameterSemantics(
            parameter_name="approved",
            parameter_type="bool",
            components=[],
            indexed=False,
        ),
    ],
)

erc721_balanceOf_function = FunctionSemantics(
    signature="0x70a08231",
    name="balanceOf",
    inputs=[ParameterSemantics(parameter_name="owner", parameter_type="address")],
    outputs=[ParameterSemantics(parameter_name="", parameter_type="uint256")],
)

erc721_ownerOf_function = FunctionSemantics(
    signature="0x6352211e",
    name="ownerOf",
    inputs=[ParameterSemantics(parameter_name="tokenId", parameter_type="uint256")],
    outputs=[ParameterSemantics(parameter_name="", parameter_type="address")],
)

erc721_ownerOf_function_transformation = {
    "__input0__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input0__)"
    )
}

erc721_transferFrom_function = FunctionSemantics(
    signature="0x23b872dd",
    name="transferFrom",
    inputs=[
        ParameterSemantics(parameter_name="from", parameter_type="address"),
        ParameterSemantics(parameter_name="to", parameter_type="address"),
        ParameterSemantics(parameter_name="tokenId", parameter_type="uint256"),
    ],
)

erc721_transferFrom_function_transformation = {
    "__input2__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input2__)"
    )
}

erc721_safeTransferFrom_function = FunctionSemantics(
    signature="0x42842e0e",
    name="safeTransferFrom",
    inputs=[
        ParameterSemantics(parameter_name="from", parameter_type="address"),
        ParameterSemantics(parameter_name="to", parameter_type="address"),
        ParameterSemantics(parameter_name="tokenId", parameter_type="uint256"),
    ],
)

erc721_safeTransferFrom_function_transformation = {
    "__input2__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input2__)"
    )
}

erc721_safeTransferFrom_with_data_function = FunctionSemantics(
    signature="0xb88d4fde",
    name="safeTransferFrom",
    inputs=[
        ParameterSemantics(parameter_name="from", parameter_type="address"),
        ParameterSemantics(parameter_name="to", parameter_type="address"),
        ParameterSemantics(parameter_name="tokenId", parameter_type="uint256"),
        ParameterSemantics(
            parameter_name="data", parameter_type="bytes", components=[], dynamic=True
        ),
    ],
)

erc721_safeTransferFrom_with_data_function_transformation = {
    "__input2__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input2__)"
    )
}

erc721_approve_function = FunctionSemantics(
    signature="0x095ea7b3",
    name="approve",
    inputs=[
        ParameterSemantics(parameter_name="operator", parameter_type="address"),
        ParameterSemantics(parameter_name="tokenId", parameter_type="uint256"),
    ],
)

erc721_approve_function_transformation = {
    "__input1__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input1__)"
    )
}

erc721_setApprovalForAll_function = FunctionSemantics(
    signature="0xa22cb465",
    name="setApprovalForAll",
    inputs=[
        ParameterSemantics(parameter_name="address", parameter_type="address"),
        ParameterSemantics(parameter_name="approved", parameter_type="bool"),
    ],
)

erc721_getApproved_function = FunctionSemantics(
    signature="0x081812fc",
    name="getApproved",
    inputs=[ParameterSemantics(parameter_name="tokenId", parameter_type="uint256")],
    outputs=[ParameterSemantics(parameter_name="", parameter_type="address")],
)

erc721_getApproved_function_transformation = {
    "__input0__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input0__)"
    )
}

erc721_isApprovedForAll_function = FunctionSemantics(
    signature="0xe985e9c5",
    name="isApprovedForAll",
    inputs=[
        ParameterSemantics(parameter_name="owner", parameter_type="address"),
        ParameterSemantics(parameter_name="operator", parameter_type="address"),
    ],
    outputs=[ParameterSemantics(parameter_name="", parameter_type="bool")],
)

ERC721_EVENTS = {
    erc721_transfer_event.signature: erc721_transfer_event,
    erc721_approval_event.signature: erc721_approval_event,
    erc721_approvalForAll_event.signature: erc721_approvalForAll_event,
}

ERC721_FUNCTIONS = {
    erc721_balanceOf_function.signature: erc721_balanceOf_function,
    erc721_ownerOf_function.signature: erc721_ownerOf_function,
    erc721_transferFrom_function.signature: erc721_transferFrom_function,
    erc721_safeTransferFrom_function.signature: erc721_safeTransferFrom_function,
    erc721_safeTransferFrom_with_data_function.signature: erc721_safeTransferFrom_with_data_function,
    erc721_approve_function.signature: erc721_approve_function,
    erc721_setApprovalForAll_function.signature: erc721_setApprovalForAll_function,
    erc721_getApproved_function.signature: erc721_getApproved_function,
    erc721_isApprovedForAll_function.signature: erc721_isApprovedForAll_function,
}

ERC721_TRANSFORMATIONS = {
    erc721_transfer_event.signature: erc721_transfer_event_transformation,
    erc721_approval_event.signature: erc721_approval_event_transformation,
    erc721_ownerOf_function.signature: erc721_ownerOf_function_transformation,
    erc721_transferFrom_function.signature: erc721_transferFrom_function_transformation,
    erc721_safeTransferFrom_function.signature: erc721_safeTransferFrom_function_transformation,
    erc721_safeTransferFrom_with_data_function.signature: erc721_safeTransferFrom_with_data_function_transformation,
    erc721_approve_function.signature: erc721_approve_function_transformation,
    erc721_getApproved_function.signature: erc721_getApproved_function_transformation,
}



from ethtx.models.objects_model import (
    BlockMetadata,
    TransactionMetadata,
    Event,
    Call,
    Transaction,
    Block,
)
from tests.models.mock import FAKE_TIME, ObjectModelMock


class TestObjectsModels:
    def test_block_metadata(self):
        bm = BlockMetadata(
            block_number=15,
            block_hash="0x1",
            timestamp=FAKE_TIME,
            parent_hash="0x",
            miner="miner",
            gas_limit=12,
            gas_used=1,
            tx_count=5,
        )

        assert bm.block_number == 15
        assert bm.block_hash == "0x1"
        assert bm.timestamp == FAKE_TIME
        assert bm.parent_hash == "0x"
        assert bm.miner == "miner"
        assert bm.gas_limit == 12
        assert bm.gas_used == 1
        assert bm.tx_count == 5
        assert bm.canonical

    def test_transaction_metadata(self):
        tm = TransactionMetadata(
            tx_hash="0x",
            block_number=1,
            gas_price=2,
            from_address="0xa",
            to_address="0xb",
            tx_index=3,
            tx_value=4,
            gas_limit=5,
            gas_used=1,
            success=False,
        )

        assert tm.tx_hash == "0x"
        assert tm.block_number == 1
        assert tm.gas_price == 2
        assert tm.from_address == "0xa"
        assert tm.to_address == "0xb"
        assert tm.tx_index == 3
        assert tm.tx_value == 4
        assert tm.gas_limit == 5
        assert tm.gas_used == 1
        assert not tm.success
        assert tm.gas_refund is None
        assert tm.return_value is None
        assert tm.exception_error is None
        assert tm.exception_error_type is None
        assert tm.revert_reason is None

    def test_event(self):
        e = Event(contract="0x", topics=["", ""], log_index=1)

        assert e.contract == "0x"
        assert e.topics == ["", ""]
        assert e.log_data is None
        assert e.log_index == 1
        assert e.call_id is None

    def test_call(self):
        c = Call(
            call_type="call",
            from_address="0xa",
            to_address="0xb",
            call_value=10,
            call_data="0x00000000000000000",
            return_value="0xeeee",
            status=True,
        )

        assert c.call_type == "call"
        assert c.from_address == "0xa"
        assert c.to_address == "0xb"
        assert c.call_value == 10
        assert c.call_gas is None
        assert c.call_data == "0x00000000000000000"
        assert c.return_value == "0xeeee"
        assert c.status
        assert c.gas_used is None
        assert c.error is None
        assert c.subcalls == []
        assert c.call_id is None
        assert c.created_address is None
        assert c.gas_refund is None
        assert c.exception_error is None
        assert c.exception_error_type is None
        assert c.revert_reason is None
        assert c.success is None

    def test_transaction(self):
        t = Transaction(
            metadata=ObjectModelMock.TRANSACTION_METADATA,
            root_call=ObjectModelMock.CALL,
            events=[
                ObjectModelMock.EVENT,
                ObjectModelMock.EVENT,
                ObjectModelMock.EVENT,
            ],
        )

        assert t.metadata == ObjectModelMock.TRANSACTION_METADATA
        assert t.root_call == ObjectModelMock.CALL
        assert t.events == [
            ObjectModelMock.EVENT,
            ObjectModelMock.EVENT,
            ObjectModelMock.EVENT,
        ]

    def test_block(self):
        t = Transaction(
            metadata=ObjectModelMock.TRANSACTION_METADATA,
            root_call=ObjectModelMock.CALL,
            events=[
                ObjectModelMock.EVENT,
                ObjectModelMock.EVENT,
                ObjectModelMock.EVENT,
            ],
        )

        b = Block(
            chain_id="mainnet",
            metadata=ObjectModelMock.BLOCK_METADATA,
            transactions=[t, t],
        )

        assert b.chain_id == "mainnet"
        assert b.metadata == ObjectModelMock.BLOCK_METADATA
        assert b.transactions == [t, t]



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


__all__ = [
    "NodeConnectionException",
    "ProcessingException",
    "InvalidTransactionHash",
    "InvalidEtherscanReturnCodeException",
    "FourByteConnectionException",
    "FourByteContentException",
    "FourByteException",
]

import json
from typing import Dict


class NodeConnectionException(Exception):
    """Node Connection Exception."""

    def __init__(self):
        super().__init__("Couldn't connect to node(s)")


class ProcessingException(Exception):
    """Processing Exception."""

    def __init__(self, msg):
        super().__init__("Exception processing: " + msg)


class InvalidTransactionHash(Exception):
    """Invalid Transaction Hash."""

    def __init__(self, tx_hash):
        super().__init__("Invalid transaction hash provided: " + tx_hash)


class InvalidEtherscanReturnCodeException(Exception):
    def __init__(self, returned_code: int, params: Dict = None):
        params_msg = " with params: " + json.dumps(params) if params else ""
        msg = f"Invalid status code for etherscan request: {returned_code} {params_msg}"
        super().__init__(msg)


class FourByteException(Exception):
    """4byte base exception class."""


class FourByteConnectionException(FourByteException):
    """4byte directory connection error."""

    def __init__(self, msg: str):
        super().__init__(f"Couldn't connect to 4byte.directory: {msg}")


class FourByteContentException(FourByteException):
    """4byte content exception. Missing output."""

    def __init__(self, status_code: int, content: bytes):
        super().__init__(
            f"Wrong response from 4byte.directory. Status code:{status_code}, content: {content}"
        )
